#include <Python.h>
#include <casa/aips.h>
#include <casa/Arrays.h>
#include <casa/Containers/Record.h>
#include <casa/Arrays/ArrayPosIter.h>
#include <tables/Tables.h>
#include <tables/Tables/BaseTable.h>
#include <tables/Tables/TableParse.h>
#include <tables/Tables/TableKeyword.h>

#include "stringifier.h"

#include <boost/python.hpp>

#include <arrayobject.h>

#include <algorithm>
#include <map>


using namespace casa;
using namespace std;
using namespace boost::python;

// exception cruft
struct column_proxy_no_such_column:
    public std::exception 
{};
struct column_proxy_unsupported_column_datatype:
    public std::exception 
{};

struct failed_to_load_Numarray_API:
    public std::exception
{};

struct failed_to_create_Numarray_object:
    public std::exception
{};

struct pyobject_is_notapybool:
    public std::exception
{};

struct incompatible_pythontype_for_columndatatype:
    public std::exception
{};

struct shape_of_subarrays_not_all_equal:
    public std::exception
{};

// mapping of casa::datatype -> python Numeric datatype
typedef std::map<casa::DataType,PyArray_TYPES>   dtmap_t;

dtmap_t make_dtmap( void )
{
    dtmap_t  rv;

    rv[TpBool]       = PyArray_INT;
    rv[TpInt]        = PyArray_INT;
    rv[TpUInt]       = PyArray_UINT;
    rv[TpString]     = PyArray_OBJECT;
    rv[TpDouble]     = PyArray_DOUBLE;
    rv[TpFloat]      = PyArray_FLOAT;
    rv[TpComplex]    = PyArray_CFLOAT;
    rv[TpDComplex]   = PyArray_CDOUBLE;
    rv[TpArrayFloat] = PyArray_OBJECT;
    return rv;
}

dtmap_t   dtmap = make_dtmap();

// some support stuff
namespace support {
    // PyBool objects don't hold a value
    // but point at the static Py_True or
    // Py_False objects. The thing they point
    // at determines the truthiness of the PyBool.
    // We need a fn which translates this pointer-to-truthiness
    // into a plain old c++ boolean.
    // Throws exception if o does not point to
    // either Py_True||Py_False
    bool pybool_ascppbool(PyObject* o) {
        bool rv;

        if( o==Py_False ) {
            rv = false;
        } else {
            if( o==Py_True ) {
                rv = true;
            } else {
                throw pyobject_is_notapybool();
            }
        }
        return rv;
    }

    // be able to translate between
    // PyComplex and PyDComplex and
    // a two-element-tuple of float/double
    // (which will allow easy translation to
    // other objects (eg casa::Complex...)
    template <typename T>
    struct mcplx {
        typedef T basic_type;

        // default: uninitialized...
        mcplx() {}
        
        // construct from real/imag of type T
        mcplx(T r, T i) {
            __bt[0] = r;
            __bt[1] = i;
        }
        
        // construct from a Py_complex,
        // doing type-casting
        // Compiler should warn if we're doing
        // something which is well.. less optimal
        mcplx(const Py_complex& pc ) {
            __bt[0] = T(pc.real);
            __bt[1] = T(pc.imag);
        }

        const T&    real( void ) const {
            return __bt[0];
        }
        const T&    imag( void ) const {
            return __bt[1];
        }

        private:
            basic_type   __bt[2];
    };

    // hmmmm python check fn's are #defines
    // and we cannot stick those in templates. Bummer.
    // have to have wrappers. Shoot.
    int pyint_check(PyObject* o) {
        return PyInt_Check(o);
    }
    int pylong_check(PyObject* o) {
        return PyLong_Check(o);
    }
    int pyfloat_check(PyObject* o) {
        return PyFloat_Check(o);
    }
    int pybool_check(PyObject* o) {
        return PyBool_Check(o);
    }
    int pycomplex_check(PyObject* o) {
        return PyComplex_Check(o);
    }
    int pystring_check(PyObject* o) {
        return PyString_Check(o);
    }
}

// sjeezus... operator== on casa::IPosition
// throws!!! who-the-fucking-hell EVER thought
// that one up?!
class JPosition:
    public casa::IPosition
{
    public:
        JPosition():
            casa::IPosition()
        {}
        JPosition(casa::uInt l):
            casa::IPosition(l)
        {}
        JPosition(casa::uInt l, casa::Int v):
            casa::IPosition(l, v)
        {}
        
        // jesus-holy-tapdancing-Christ!
        // Even bloody-operator=() will throw?!?!
        const JPosition& operator=( const JPosition& other ) {
            this->resize(0);
            this->casa::IPosition::operator=(other);
            return *this;
        }
        const JPosition& operator=( const casa::IPosition& other ) {
            this->resize(0);
            this->casa::IPosition::operator=(other);
            return *this;
        }
};

// the versions of operator==
bool operator==(const JPosition& l, const JPosition& r) {
    return l.casa::IPosition::isEqual(r);
}
bool operator==(const JPosition& l, const casa::IPosition& r) {
    return l.casa::IPosition::isEqual(r);
}
bool operator==(const casa::IPosition& l, const JPosition& r) {
    return l.isEqual(r);
}
// id. for operator !=
bool operator!=(const JPosition& l, const JPosition& r) {
    return !(l==r);
}
bool operator!=(const JPosition& l, const casa::IPosition& r) {
    return !(l==r);
}
bool operator!=(const casa::IPosition& l, const JPosition& r) {
    return !(l==r);
}

namespace sequencesupport {
    //using casa::IPosition;

    // same as in support::py*_check to
    // translate #defines into 'real'
    // functions but now for sequence types
    int pynumarray_check(PyObject* o) {
        return PyArray_Check(o);
    }
    int pytuple_check(PyObject* o) {
        return PyTuple_Check(o);
    }
    int pylist_check(PyObject* o) {
        return PyList_Check(o);
    }
    int pysequence_check(PyObject* o) {
        return PySequence_Check(o);
    }

    // size functions
    int pylist_size(PyObject* o) {
        return PyList_Size(o);
    }
    int pytuple_size(PyObject* o ) {
        return PyTuple_Size(o);
    }

    // element access
    PyObject* pylist_getitem(PyObject* o, int idx) {
        return PyList_GetItem(o, idx);
    }
    PyObject* pytuple_getitem(PyObject* o, int idx) {
        return PyTuple_GetItem(o, idx);
    }

    // we must be able to get the shape of a sequence
    // the shaper interface must take a PyObject* and
    // get a casa::IPosition from that.
    JPosition numarray_shape( PyObject* o );
    JPosition tuple_shape( PyObject* o );
    JPosition list_shape( PyObject* o );
    // if the object is a list of scalars,
    // that's acceptable as well...
    JPosition scalar_shape( PyObject* o );

    typedef JPosition (*sfnptr)(PyObject*);
    template <sfnptr S1, sfnptr S2, sfnptr S3, sfnptr S4>
    struct shaper {
        static JPosition get_shape(PyObject* o) {
            JPosition              r;
            static const JPosition dummy;
            // check all alternatives
            if( (r=S1(o))==dummy ) {
                if( (r=S2(o))==dummy ) {
                    if( (r=S3(o))==dummy ) {
                        r=S4(o);
                    }
                }
            }
            return r;
        }
    };
    typedef shaper<scalar_shape, numarray_shape, tuple_shape, list_shape> Shaper;
   
    // the real shapers
    JPosition numarray_shape( PyObject* o ) {
        JPosition rv;

        // if the PyArray API not loaded, then load it!
        if( import_array()<0 ) {
            throw failed_to_load_Numarray_API();
        }
        if( pynumarray_check(o) ) {
            int             pyidx;
            unsigned int    casaidx;
            PyArrayObject*  aoptr = (PyArrayObject*)o;
            
            rv = JPosition(aoptr->nd);
            for( pyidx=0, casaidx=0; pyidx<aoptr->nd; ++pyidx, ++casaidx ) {
                rv[casaidx] = aoptr->dimensions[pyidx];
            }
        }
        return rv;
    }

    // function template for getting the shape of a PyObject
    // which supports
    //   1) checking for type
    //   2) getting 1D-size 
    //   3) element access
    // will recursively check all dims
    template <int (*checkfn)(PyObject*),
              int (*sizefn)(PyObject*),
              PyObject* (*getfn)(PyObject*, int)>
    struct pyobj_shape {
        // the static fn
        static JPosition get_shape( PyObject* o ) {
            cerr << "pyobj_shape::get_shape(" << hex << (unsigned int)o << dec << ")" << endl;
            JPosition rv;
            JPosition myshape;
            // if not the type we need
            if( !checkfn(o) )
                return JPosition();

            cerr << "pyobj_shape::checkfn was Ok" << endl;
            // we are a tuple so our basic shape is
            // a 1D array of length PyTuple_Size
            myshape = JPosition(1, sizefn(o));
        
            // check if we need to check anything at all
            if( myshape[0]==0 )
                return myshape;

            cerr << "pyobj_shape::myshape[0]=" << myshape[0] << endl;
            // ok there are elements to check. Assert
            // that all elements have the same shape.
            int          idx;
            PyObject*    ob0;
            JPosition    shp0;

            ob0  = getfn(o, 0);
            cerr << "pyobj_shape::got ob0=" << hex << (unsigned int)ob0 << dec << endl;
            shp0 = Shaper::get_shape(ob0);
            Py_DECREF(ob0);
            cerr << "pyobj_shape::ob0's shape=" << shp0 << endl;
            for( idx=1; idx<myshape[0]; idx++ ) {
                cerr << "pyobj_shape::checking obj #" << idx << endl;
                PyObject*  optr;
                JPosition  shpn;

                optr = getfn(o, idx);
                shpn = Shaper::get_shape(optr);
                Py_DECREF(optr);
                if( shpn!=shp0 )
                    break;
            }
            if( idx!=myshape[0] ) {
                return JPosition();
                //throw shape_of_subarrays_not_all_equal();    
            }
            // Now construct the final return-value:
            // we are a shp0.ndim()+1 array
            // with dim[0]==myshape[0] and
            // the other dims are shp0
            myshape.append( shp0 );
            cerr << "pyobj_shape::subshapes verified, total=" << myshape << endl;
            return myshape;
        }
    };

    // the functions basically use the functiontemplates
    JPosition tuple_shape(PyObject* o) {
        return pyobj_shape<pytuple_check, pytuple_size, pytuple_getitem>::get_shape(o);
    }
    JPosition list_shape(PyObject* o) {
        return pyobj_shape<pylist_check, pylist_size, pylist_getitem>::get_shape(o);
    }

    // for acceptable scalartypes
    typedef int (*cfn)(PyObject*);
    template <cfn S1, cfn S2, cfn S3, cfn S4, cfn S5, cfn S6>
    struct is_scalar {
        static int do_test(PyObject* o) {
            int  r;
            if( (r=S1(o))==0 ) {
                if( (r=S2(o))==0 ) {
                    if( (r=S3(o))==0 ) {
                        if( (r=S4(o))==0 ) {
                            r=S5(o);
                        }
                    }
                }
            }
            return r;
        }
    };
    typedef is_scalar<support::pyint_check,
                      support::pylong_check,
                      support::pystring_check,
                      support::pyfloat_check,
                      support::pybool_check,
                      support::pycomplex_check> Scalar;

    // the scalar-shaper uses
    // this asserter, and asserts that
    //  1) the pyobject is a certain (sequence) type
    //  2) all of the elements are scalars
    template <int (*checkfn)(PyObject*),
              int (*sizefn)(PyObject*),
              PyObject* (*getfn)(PyObject*, int)>
    JPosition scalar_asserter(PyObject* o) {
        if( !checkfn(o) )
            return JPosition();
        cerr << "scalar_asserter: begins" << endl;
        JPosition  myshape( 1, sizefn(o) );
        
        if( myshape[0]==0 )
            return myshape;
        int   idx; 
        cerr << "scalar_asserter: checking " << myshape[0] << " elements" << endl;
        for(idx=0; idx<myshape[0]; idx++) {
            int        sclr;
            PyObject*  optr;
            optr = getfn(o, idx);
            sclr = Scalar::do_test(optr);
            Py_DECREF(optr);
            if( !sclr )
               break;
        }
        if( idx!=myshape[0] )
            myshape = JPosition();
        cerr << "scalar_asserter: myshape=" << myshape << endl;
        return myshape;
    }
    
    JPosition scalar_shape(PyObject* o) {
        JPosition              r;
        static const JPosition empty;
        if( (r=scalar_asserter<pytuple_check, pytuple_size, pytuple_getitem>(o))==empty ) {
            r=scalar_asserter<pylist_check, pylist_size, pylist_getitem>(o);
        }
        return r;
    }
}
// stick some conversion (warp) functors
// into their own namepsace.
// These functors can be used to transform
// something into a compile-time known returntype.
// These 'warpers' are used to transform
// the elements of casa arrays into their
// numarray supported equivalent layout.
// Typically, Python numarray objects are
// plain-old-C-style arrays and for some
// casa datatypes we MUST make sure that
// the layout in the numarray object adheres
// to the layout numarray expects.
namespace warp {
    template <typename T>
    struct identity {
        typedef T  returntype;
        T operator()( const T& t ) const {
            return t;
        }

        // warp a number of objects in one go
        // Since this is the *identity* warper
        // it assumes T is POD [plain old data]
        // hence that bitwise copies are Ok
        // For objects which must use 'operator='
        // to work properly, we must implement
        // a different warper. For now it's not
        // necessary so...
        void operator()(T* dst, const T* src, unsigned int n) const {
            ::memcpy(dst, src, n*sizeof(T));
            return;
        }
    };

    // casa::Bool is represented as int  (PyArray_INT)
    // so we warp a casa::Bool into that. The identity
    // relation (warp::identity) didn't apparently hold
    // very well for this type...
    template <typename T>
    struct casabool2int {
        typedef int   returntype;
        
        // only support warping of casa::Bool for now...
        int operator()( const casa::Bool& b ) const {
            return (b==True)?(1):(0);
        }

        // a number of bools in one go
        void operator()(int* dst, const casa::Bool* src, unsigned int n) const {
            for( unsigned int i=0; i<n; ++i ) {
                *dst++ = (*this)(*src++);
            }
        }
    };

    // the template parameter here is unused but at least
    // now this thang can be used as template template
    // parameter
    template <typename T=int>
    struct ToPyStr {
        typedef PyObject* returntype;
        PyObject* operator()( const casa::String& t ) const {
            return PyString_FromString(t.chars());
        }

        // nothing else to do but to loop over all items...
        void operator()(returntype* dst, const casa::String* src, unsigned int n) const {
            for( unsigned int i=0; i<n; ++i ) {
                *dst++ = (*this)(*src++);
            }
            return;
        }
    };

    // for casa::Complex and casa::DComplex

    // this warper ignores the template argument
    // and converts whatever is passed to it
    // to an 'array of 2 floats'
    template <typename T>
    struct ToPyComplex {
        typedef float basic_type;
        typedef struct {
            basic_type   re;
            basic_type   im;
        } returntype;

        // Support anything that has .real() and .imag()
        // which are convertible to 'float'
        template <typename U>
        returntype operator()( const U& c ) const {
            returntype    rv;

            rv.re = basic_type(c.real());
            rv.im = basic_type(c.imag());
            return rv;
        }

        // first, the generic, slow implementation
        // which supports converting an array of
        // U's which have .real() and .imag()
        // which are convertible to float.
        template <typename U>
        void operator()(returntype* dst, const U* src, unsigned int n) const {
            for( unsigned int i=0; i<n; ++i ) {
                *dst++ = (*this)(*src++);
            }
            return;
        }

        // specialization for casa::Complex -> these are
        // in layout identical to PyComplex so we can
        // optimize now and just do a plain memcpy!
        void operator()(returntype* dst, const casa::Complex* src, unsigned int n) const {
            ::memcpy(dst, src, n*sizeof(returntype));
            return;
        }
    };

    template <typename T>
    struct ToPyDComplex {
        typedef double basic_type;
        typedef struct {
            basic_type   re;
            basic_type   im;
        } returntype;

        // Support anything that has .real() and .imag()
        // which are convertible to 'double'
        template <typename U>
        returntype operator()( const U& c ) const {
            returntype    rv;

            rv.re = basic_type(c.real());
            rv.im = basic_type(c.imag());
            return rv;
        }

        // first, the generic, slow implementation
        // which supports converting an array of
        // U's which have .real() and .imag()
        // which are convertible to double.
        template <typename U>
        void operator()(returntype* dst, const U* src, unsigned int n) const {
            for( unsigned int i=0; i<n; ++i ) {
                *dst++ = (*this)(*src++);
            }
            return;
        }

        // specialization for casa::DComplex -> these are
        // in layout identical to PyDComplex so we can
        // optimize now and just do a plain memcpy!
        void operator()(returntype* dst, const casa::DComplex* src, unsigned int n) const {
            ::memcpy(dst, src, n*sizeof(returntype));
            return;
        }
    };


    // section of warpers which take a PyObject
    // and transform into POD datatype 'T'.

    // define functor template for fn which
    // checks/extracts a basic type from
    // a python object (exposes the 'do_convert(T&, PyObject*)'
    // interface
    //
    // This should only go for
    //   long + PyInt_AsLong
    //   long + PyLong_AsLong
    //   double + PyFloat_AsDouble
    //   bool + support::pybool_ascppbool
    //   Py_complex + PyComplex_AsCComplex
    //   char* + PyString_AsString
    template <typename T, int (*checkfn)(PyObject*), T (*convfn)(PyObject*)>
    struct fromPythonBasic {
        typedef T return_type;
        static bool do_convert(T& t, PyObject* o ) {
            bool rv( false );
            if( checkfn(o) ) {
               t  = T(convfn(o));
               rv = true;
            }
            return rv;
        }
    };

    // this one accepts all calls and always
    // returns false. Useful for non-existant
    // alternatives. We must be prepared to 
    // be able to accept alternatives but 
    // sometimes there just aren't any.
    // For those cases, use this'un
    template <typename T=int>
    struct no_alternative {
        typedef T return_type;

        template <typename U>
        static bool do_convert( U& , PyObject* ) {
            return false;
        }
    };

    // this is a template which allows you
    // to specify alternate conversions
    // if you go from Python->plain-old-c++
    template <typename T,
              typename C1 = no_alternative<T>,
              typename C2 = no_alternative<T>,
              typename C3 = no_alternative<T> >
    struct fromPython {
        typedef T  return_type;

        // also exposes the 'do_convert(T&, PyObject*)'
        // method but tries the alternatives in sequence
        // until succesfull or alternatives exhausted
        static bool do_convert(T& t, PyObject* o ) {
            bool                     rv;
            typename C1::return_type c1;

            if( (rv=C1::do_convert(c1, o))==true ) {
                t  = T(c1);
            } else {
                typename C2::return_type  c2;
                if( (rv=C2::do_convert(c2, o))==true ) {
                    t = T(c2);
                } else {
                    typename C3::return_type  c3;
                    if( (rv=C3::do_convert(c3, o))==true ) {
                        t = T(c3);
                    }
                }
            }
            return rv;
        }
    };
    // short-hand typedefs for basic conversions
    // read something as "_ial" to mean "PyInt-as-long"
    // or "_laul" to mean "PyLong-as-unsigned-long"
    typedef fromPythonBasic<long, support::pyint_check, PyInt_AsLong> fpb_ial;
    typedef fromPythonBasic<long, support::pylong_check, PyLong_AsLong> fpb_lal;
    typedef fromPythonBasic<unsigned long, support::pyint_check, PyInt_AsUnsignedLongMask> fpb_iaul;
    typedef fromPythonBasic<unsigned long, support::pylong_check, PyLong_AsUnsignedLong> fpb_laul;
    typedef fromPythonBasic<double, support::pyfloat_check, PyFloat_AsDouble> fpb_dad;
    typedef fromPythonBasic<bool, support::pybool_check, support::pybool_ascppbool> fpb_bab;
    typedef fromPythonBasic<Py_complex, support::pycomplex_check, PyComplex_AsCComplex> fpb_cac;
    typedef fromPythonBasic<char*, support::pystring_check,PyString_AsString>  fpb_sas;

    // Now, from the basic conversions, we may build conversion chains/alternatives.
    // We may accept different source python types for the same destination
    // C/C++ type
    // Eg: for python -> int we support both PyInt & PyLong
    typedef fromPython<int,          fpb_ial, fpb_lal>          as_int;
    typedef fromPython<unsigned int, fpb_iaul, fpb_laul>        as_uint;
    // for float/double we accept PyFloat but also PyInt and PyLong
    typedef fromPython<float,        fpb_dad, fpb_ial, fpb_lal> as_float;
    typedef fromPython<double,       fpb_dad, fpb_ial, fpb_lal> as_double;

    // for bool we support PyBool, PyInt and PyLong
    typedef fromPython<bool,        fpb_bab, fpb_ial, fpb_lal>  as_bool;

    // string
    typedef fromPython<std::string, fpb_sas>                    as_string;

    // For complex we go from PyObject*[PyComplex] -> Py_complex (C-struct with 2 doubles)
    // to intermediate type 'mcplx<T>' which allows converting from
    // Py_complex to any other complex type (int? float? double?)
    // The mcplx<T> has '.real()' and '.imag()' methods for extracting.. well, real+imag.
    // The mcplx<T> has a constructor from Py_complex which (attempts) the
    //  conversion from double to T.
    typedef fromPython<support::mcplx<float>, fpb_cac>          as_fcomplex;
    typedef fromPython<support::mcplx<double>, fpb_cac>         as_dcomplex;
}


//
// Transform casa::Array to a numarray
// 
template <template <typename> class Transform=warp::identity>
struct arraytransformer {
    template <typename T>
    PyObject* operator()( const Array<T>& t ) const
    {
        IPosition                           shp( t.shape() );
        IPosition                           newshp( shp.nelements() );
        Transform<T>                        warper;
        PyArrayObject*                      po;
        dtmap_t::const_iterator             dstdt;
        typename Transform<T>::returntype*  dst;

        // if we don't recognize the type, tell 'm so!
        // The datatypemap tells us, given the casa datatype,
        // what the corresponding Numarray datatype would be
        if( (dstdt=dtmap.find(whatType((T*)0)))==dtmap.end() )
        {
            // if this occurs, maybe you need to 
            // add an entry to the 'dtmap' variable....
            // browse this file and look for it.
            ostringstream s;
            s << __FILE__ << ":" << __LINE__ << " - arraytransformer<> -> Unhandled datatype "
                 << whatType((T*)0);
            cerr << s.str() << endl;
            PyErr_SetString(PyExc_ValueError,s.str().c_str());
            return NULL;
        }

        // ok, so we recognized the type, only the array's empty...
        if( shp.product()==0 ) {
            Py_INCREF(Py_None);
            return Py_None;
        }

        // if the PyArray API not loaded, then load it!
        if( import_array()<0 ) {
            throw failed_to_load_Numarray_API();
        }
#if 0
        if( !PyArray_API ) {
            import_array();
        }
        if( PyArray_API==0 ) {
            throw failed_to_load_Numarray_API();
        }
#endif

        // create the array from the dimensions
        po = (PyArrayObject*)PyArray_FromDims(shp.nelements(),
                                              (int*)shp.storage(),
                                              dstdt->second);

        if( po==0 ) {
            PyErr_SetString(PyExc_MemoryError, "Failed to allocate Numarray-object");
            return NULL;
        }

        // do a cast to destination type
        dst = (typename Transform<T>::returntype*)po->data;

        // do it like this to be able to *transform*
        // between different types...
        // Eg when transforming an array of strings;
        // each string has to be transformed to
        // a PyObject in that case...
        //
        // So we first get a contiguous array
        // out of the source array such that
        // we can loop as fast as possible
        // over the elements.
        //
        // The 'warper' will be called upon
        // to do the transformation.
        // There may be specializations
        // which do a memcpy, others use 
        // a for-loop with operator=()
        // [sometimes you just have to do it like
        //  that...]
        const T*   src;
        casa::Bool del;

        src = t.getStorage(del);
        warper(dst, src, shp.product());
        t.freeStorage(src, del);

        // copy the strides to fix
        // array-indexing
        // numarray strides are in units of bytes,
        // casa steps are in units of 'items'
        // so must multiply the casa strides by
        // the itemsize
        for( unsigned int k=0; k<t.steps().nelements(); ++k )
            po->strides[k] = t.steps()[k] * sizeof(typename Transform<T>::returntype);

        return (PyObject*)po;
    }
};


// convert all the arrayvalues
struct tostdstring
{
    std::string operator()( const casa::String& cs ) const 
    {
        return std::string(cs.chars());
    }
};


// Record-field accessor, gets the (scalar) valued field
// into the variable 'value'
template <typename T>
struct rfaccessor_t {
    rfaccessor_t(const casa::RecordInterface& r, const RecordFieldId& fi) {
        r.get(fi, value);
    }

    T   value;
};

// id. as above, now for arrayvalued fields
// into the variable 'value'
template <typename T>
struct rfaaccessor_t {
    rfaaccessor_t(const casa::RecordInterface& r, const RecordFieldId& fi) {
        r.get(fi, value);
    }

    casa::Array<T>   value;
};

// function template for converting casa (scalar) valued RecordFields into
// a PyObject, using an intermediate conversion to allow translation from
// (potentially not supported by Python)casa types to
// python supported built-in types.
// E.g. the actual conversion chain of 
//      casa::Bool -> PyObject*
// goes like:
//      casa::Bool -> built-in type 'long' -> PyBool_FromLong -> PyObject*
template <typename T, typename U, PyObject* (*fptr)(U)>
struct gen {
    static PyObject* do_gen(const casa::RecordInterface& r, Int i) {
        rfaccessor_t<T>  rfa(r, i);
        return fptr(U(rfa.value));
    }
};

// id. for array-valued thingies (hence the 'a' suffix)
// Specify a 'Transform' template template parameter;
// this class will be used to transform individual array
// elements to a corresponding python-numarray-type
template <typename T, template <typename> class Transform = warp::identity>
struct gena {
    static PyObject* do_gen(const casa::RecordInterface& r, Int i) {
        rfaaccessor_t<T>  rfa(r, i);
        return arraytransformer<Transform>()(rfa.value);
    }
};

// function pointer which defines an interface which 
// takes a "RecordInterface" and a fieldnumber
// and warps it into a PyObject*
typedef PyObject* (*genfunc)(const casa::RecordInterface&, Int num);

// map casa datatype to a pointer-to-function so the system
// knows how to transform a casa datatype into a PyObject
typedef map<casa::DataType,genfunc> genmap_t;


PyObject* make_pystr( const casa::String& s ) {
    return PyString_FromString(s.chars());
}

// because of 'circular' dependency, or, recursiveness if you will,
// the function which turns a (Table)Record into a PyDict must be able
// to find itself in the global 'genmap_t' variable, but from the
// function we must be able to reference the global genmap_t variable,
// decide to have a placeholder here which we can stick into the map.
PyObject* make_pydict_fromsubrecord(casa::RecordInterface const&, Int num);

PyObject* make_pyobject_fromsubtable(casa::RecordInterface const&, Int num);


// fill in the mapping
genmap_t make_genmap()
{
    genmap_t   rv;

    rv[TpBool] = &gen<casa::Bool,long,PyBool_FromLong>::do_gen;
    rv[TpInt] =  &gen<Int,long,PyInt_FromLong>::do_gen;
    rv[TpUInt] = &gen<uInt,long,PyInt_FromLong>::do_gen;
    rv[TpFloat] = &gen<Float,double,PyFloat_FromDouble>::do_gen;
    rv[TpDouble] = &gen<Double,double,PyFloat_FromDouble>::do_gen;
    rv[TpString] = &gen<casa::String,const casa::String&,make_pystr>::do_gen;

    rv[TpRecord] = &make_pydict_fromsubrecord;
    rv[TpTable]  = &make_pyobject_fromsubtable;

    rv[TpArrayBool] = &gena<casa::Bool>::do_gen;
    rv[TpArrayInt]  = &gena<Int>::do_gen;
    rv[TpArrayFloat] = &gena<Float>::do_gen;
    rv[TpArrayDouble]= &gena<Double>::do_gen;
    rv[TpArrayComplex] = &gena<Complex, warp::ToPyComplex>::do_gen;
    rv[TpArrayDComplex] = &gena<DComplex, warp::ToPyDComplex>::do_gen;
    rv[TpArrayString]   = &gena<String, warp::ToPyStr>::do_gen;

    return rv;
}
genmap_t   genmap = make_genmap();



namespace boost { namespace python {
    // convert  casa String to PyString
    template <>
    struct to_python_value<casa::String const&> :
        detail::builtin_to_python
    {
        inline PyObject* operator()(casa::String const& cs) const
        {
            return ::PyString_FromString(cs.chars());
        }
    };

    // convert casa::Record to PyDict
    template <>
    struct to_python_value<casa::RecordInterface const&>:
        detail::builtin_to_python
    {
        inline PyObject* operator()( casa::RecordInterface const& ri) const
        {
            PyDictObject*   dict = (PyDictObject*)PyDict_New();

            if( !dict ) {
                return NULL;
            }

            // loop over all recordfields and transform them into
            // appropriate PythonObject* 's
            for( uInt fld=0; fld<ri.nfields(); fld++ )
            {
                genmap_t::iterator curgen( genmap.find(ri.type((Int)fld)) );

                if( curgen!=genmap.end() ) {
                    PyDict_SetItem(  (PyObject*)dict,
                                     warp::ToPyStr<>()( ri.name((Int)fld) ),
                                     curgen->second( ri, (Int)fld ) );
                } else {
                    cout << "Unknown fieldtype for " << ri.name((Int)fld)
                         << " (type=" << ri.type((Int)fld) << ")" << endl;
                }
            }
            return (PyObject*)dict;
        }
    };

}} // boost::python



// And do it the other way around ->
// Convert from python string to
// casa::String
struct CasaStringFromPythonString {
    CasaStringFromPythonString()
    {
        converter::registry::push_back(&convertible, &construct, type_id<String>() );
    }
    static void* convertible( PyObject* ptr )
    {
        if( PyString_Check(ptr)==0 )
        {
            return 0;
        }
        return ptr;
    }
    static void construct( ::PyObject* pyptr,
            boost::python::converter::rvalue_from_python_stage1_data* data )
    {
        const char* value = ::PyString_AsString(pyptr);

        if( value==0 )
        {
            boost::python::throw_error_already_set();
        }
        void* storage =
            ((boost::python::converter::rvalue_from_python_storage<casa::String>*)data)->storage.bytes;
        new (storage) casa::String(value);
        data->convertible = storage;
    }
};


// access a value in row 'r' in a scalar column 'c'
// in table 't'.
// Put the result into 'value'
template <typename T>
struct scolaccessor {
    scolaccessor(const Table& t, const std::string& c, uInt r)
    {
        ROScalarColumn<T>(t, c).get(r, value);
    }

    T            value;
};

// Give access to a (scalar)column such that we can update a value
// at row 'r'. This may sound weird but at least you can't
// influence the value in a different row after you've created
// this baby
template <typename T>
struct scolputter {
    scolputter(Table& tab, const std::string& col, uInt row, const T& val ) {
        ScalarColumn<T>    scol(tab, col);
        scol.put(row, val);
    }
};

// id. but now for an ArrayColumn
template <typename T>
struct acolaccessor {
    acolaccessor(const Table& t, const std::string& c, uInt r)
    {
        ROArrayColumn<T>(t, c).get(r, value);
    }

    Array<T>       value;
};

// Take all the values in column 'c' in table 't'
// and put the result in 'values'
// Defaults to ROScalarColumn + Vector; reading
// all the scalar values in a column into a vector.
// Can easily be used to get all the values in an
// array-valued column by changing the template
// parameters to
// ROArrayColumn + Array ....
// (big hoooorraaaaay up c++'s ass!)
template <typename T,
          template <typename> class colType = ROScalarColumn,
          template <typename> class arrType = Vector >
struct colaccessor2 {
    colaccessor2(const Table& t, const std::string& c)
    {
        colType<T>(t, c).getColumn(values, True);
    }
    arrType<T>   values;
};



template <typename T>
struct ScalarFromPython {
    ScalarFromPython(PyObject* o) {
       if( T::do_convert(value, o)==false ) {
           throw incompatible_pythontype_for_columndatatype();
       }
    }
    typename T::return_type  value;
};

// proxy for a column-object
//   Easier to keep state 
struct ColumnProxy {

    ColumnProxy( const Table& t, const std::string& col ):
        tab(t),
        colName(col)
    {
        if( !tab.tableDesc().isColumn(colName) )
        {
            throw column_proxy_no_such_column();
        }
        cd = t.tableDesc().columnDesc(colName);
    }

    unsigned int length( void ) const {
        return tab.nrow();
    }

    // access the cell at row 'idx' in the column
    // supports Python indexing scheme where
    // negative indices count backward from the end...
    PyObject* getCell( int idx )
    {
        // If idx out of range, return None
        // but first see if we must adjust the idx
        if( idx<0 ) {
            idx += (int)tab.nrow();
        }
        // if after adjustment, idx out of range
        // throw up
        if( idx<0 || idx>=tab.nrow() ) {
            PyErr_SetString(PyExc_StopIteration, "");
            return NULL;
        }

        switch( cd.trueDataType() )
        {
            case TpBool:
                {
                    scolaccessor<casa::Bool>    b(tab, colName, idx);
                    return PyInt_FromLong( ((b.value==True)?(1L):(0L)) );
                }
                break;
            case TpArrayBool:
                {
                    const acolaccessor<casa::Bool>  d(tab, colName, idx);
                    return arraytransformer<warp::casabool2int>()( d.value );
                }
                break;
            case TpInt:
                {
                    scolaccessor<Int>    b(tab, colName, idx);
                    return PyInt_FromLong( (long)b.value );
                }
                break;
            case TpArrayInt:
                {
                    const acolaccessor<Int>  d(tab, colName, idx);
                    return arraytransformer<>()( d.value );
                }
                break;

            case TpString:
                {
                    scolaccessor<String>  s(tab, colName, idx);
                    return PyString_FromString(s.value.chars());
                }
                break;
            case TpArrayString:
                {
                    const acolaccessor<String>  d(tab, colName, idx);
                    return arraytransformer<warp::ToPyStr>()( d.value );
                }
                break;

            case TpFloat:
                {
                    scolaccessor<Float>  s(tab, colName, idx);
                    return PyFloat_FromDouble((double)s.value);
                }
                break;
            case TpArrayFloat:
                {
                    const acolaccessor<Float>  d(tab, colName, idx);
                    return arraytransformer<>()( d.value );
                }
                break;

            case TpDouble:
                {
                    scolaccessor<Double>  s(tab, colName, idx);
                    return PyFloat_FromDouble(s.value);
                }
                break;
            case TpArrayDouble:
                {
                    const acolaccessor<Double>  d(tab, colName, idx);
                    return arraytransformer<>()( d.value );
                }
                break;

            case TpComplex:
                {
                    scolaccessor<Complex>  s(tab, colName, idx);
                    return PyComplex_FromDoubles(s.value.real(), s.value.imag());
                }
                break;
            case TpArrayComplex:
                {
                    const acolaccessor<Complex>  d(tab, colName, idx);
                    return arraytransformer<warp::ToPyComplex>()( d.value );
                }
                break;

            case TpDComplex:
                {
                    scolaccessor<DComplex>  s(tab, colName, idx);
                    return PyComplex_FromDoubles(s.value.real(), s.value.imag());
                }
                break;
            case TpArrayDComplex:
                {
                    const acolaccessor<DComplex>  d(tab, colName, idx);
                    return arraytransformer<warp::ToPyDComplex>()( d.value );
                }
                break;

            default:
                break;
        }
        // Unsupported type
        throw column_proxy_unsupported_column_datatype();
    }

    // Now set the value of a given cell in row 'r'
    bool putCell( int r, PyObject* o ) {
        // If idx out of range, return None
        // but first see if we must adjust the idx
        if( r<0 ) {
            r += (int)tab.nrow();
        }
        // if after adjustment, idx out of range
        // throw up
        if( r<0 || r>=tab.nrow() ) {
            PyErr_SetString(PyExc_StopIteration, "");
            return false;
        }

        // note: the 'ScalarFromPython' c'tors throw up
        // if the conversion fails
        switch( cd.trueDataType() )
        {
            case TpBool:
                {
                    ScalarFromPython<warp::as_bool>  w(o);
                    scolputter<casa::Bool>(tab, colName, r, w.value);
                }
                break;

            case TpString:
                {
                    ScalarFromPython<warp::as_string>  w(o);
                    scolputter<casa::String>(tab, colName, r, w.value);
                }
                break;

            case TpInt:
                {
                    ScalarFromPython<warp::as_int>  w(o);
                    scolputter<casa::Int>(tab, colName, r, w.value);
                }
                break;
            case TpFloat:
                {
                    ScalarFromPython<warp::as_float>  w(o);
                    scolputter<casa::Float>(tab, colName, r, w.value);
                }
                break;
            case TpDouble:
                {
                    ScalarFromPython<warp::as_double>  w(o);
                    scolputter<casa::Double>(tab, colName, r, w.value);
                }
                break;
            case TpComplex:
                {
                    ScalarFromPython<warp::as_fcomplex>  w(o);
                    scolputter<casa::Complex>(tab, colName, r, Complex(w.value.real(),
                                                                       w.value.imag()));
                }
                break;
            case TpDComplex:
                {
                    ScalarFromPython<warp::as_dcomplex>  w(o);
                    scolputter<casa::DComplex>(tab, colName, r, DComplex(w.value.real(),
                                                                         w.value.imag()));
                }
                break;
            // Array Stuff
#if 0
            case TpArrayFloat:
                {
                }
                break;
#endif                
            default:
                // Unsupported type
                throw column_proxy_unsupported_column_datatype();
                break;
        }
        // if we arrive here it means that no exceptions have been
        // thrown, ie stuff just worked!
        // flush the table to disk
        tab.flush(True);
        return true;
    }

    // get all the values in the column
    PyObject* get( void ) {
        switch( cd.trueDataType() )
        {
            case TpBool:
                {
                    const colaccessor2<casa::Bool>    b(tab, colName);
                    return arraytransformer<warp::casabool2int>()( b.values );
                }
                break;
            case TpArrayBool:
                {
                    const colaccessor2<casa::Bool,ROArrayColumn,Array>   b(tab, colName);
                    return arraytransformer<warp::casabool2int>()( b.values );
                }
                break;

            case TpInt:
                {
                    const colaccessor2<Int>    b(tab, colName);
                    return arraytransformer<>()( b.values );
                }
                break;
            case TpArrayInt:
                {
                    const colaccessor2<Int,ROArrayColumn,Array>   b(tab, colName);
                    return arraytransformer<>()( b.values );
                }
                break;

            case TpFloat:
                {
                    const colaccessor2<Float>    b(tab, colName);
                    return arraytransformer<>()( b.values );
                }
                break;
            case TpArrayFloat:
                {
                    const colaccessor2<Float,ROArrayColumn, Array>    b(tab, colName);
                    return arraytransformer<>()( b.values );
                }
                break;

            case TpDouble:
                {
                    const colaccessor2<Double>    b(tab, colName);
                    return arraytransformer<>()( b.values );
                }
                break;
            case TpArrayDouble:
                {
                    const colaccessor2<Double,ROArrayColumn, Array>    b(tab, colName);
                    return arraytransformer<>()( b.values );
                }
                break;

            case TpString:
                {
                    colaccessor2<String>     s(tab, colName);
                    return arraytransformer<warp::ToPyStr>()( s.values );
                }
                break;

            case TpArrayComplex:
                {
                    const colaccessor2<Complex, ROArrayColumn, Array>  d(tab, colName);
                    return arraytransformer<warp::ToPyComplex>()( d.values );
                }
                break;
            case TpArrayDComplex:
                {
                    const colaccessor2<DComplex, ROArrayColumn, Array>  d(tab, colName);
                    return arraytransformer<warp::ToPyDComplex>()( d.values );
                }
                break;
        }
        // Unsupported type
        throw column_proxy_unsupported_column_datatype();
    }


    // get the column keywords
    PyObject* keywordSet( void ) 
    {
        return boost::python::to_python_value<RecordInterface const&>()( cd.keywordSet() );
    }

    Table       tab;
    ColumnDesc  cd;
    std::string colName;

    private:
        ColumnProxy();
};

std::ostream& operator<<( std::ostream& os, const ColumnProxy& cp )
{
    return os << cp.cd;
}


//
// Proxy for a casa Table
//    very thin wrapper
//    but makes it easy for us
//    and Boost.Python to bind
//    only specific members and/or
//    do some processing before/after
//    the actual call to the implementation
struct TableProxy {

    // default c'tor
    TableProxy():
        tab(),
        ra( tab )
    {
    }

    // do implement the copy c'tor ->
    // this'un gets called by the BOOST stuff
    // and we must take precautions to 
    // make sure the (refcounted) table
    // and ROTableRow are properly initialized...
    // 
    // It certainly isn't valid to "just copy
    // other.tab and other.ra across" even though
    // both objects support copy construction.
    //
    // The problem is with the ROTableRow, it is,
    // when it is constructed, associated with a
    // specific table. The ROTableRow in a TableProxy
    // should always be 'associated' with the table
    // object in the same object.
    //
    // In the example below, without a "proper" (read: correct)
    // copy c'tor, after executing these statements,
    // TableProxy b's ROTableRow member references
    // TableProxy a's Table and *not* b's Table,
    // which messes up reference counting [which happens
    // inside the Tables] and hence cleaning up of
    // tempfiles etc.
    //
    // TableProxy    a;
    // TableProxy    b(a);
    //
    // A 'correct' copy constructor is implemented below,
    // we can just copy construct our table with other's table
    // but we must explicitly associate our ROTableRow with
    // *our* table-object
    TableProxy( const TableProxy& other ):
        tab( other.tab ),
        ra( tab )
    {
    }
    const TableProxy& operator=( const TableProxy& other )
    {
        if( this!=&other )
        {
            tab = other.tab;
            ra  = ROTableRow(tab);
        }
        return *this;
    }

    TableProxy( const Table& t ):
        tab( t ),
        ra( tab )
    {
    }

    // default to 'read-only'. If ro==false => opened with
    // Table::Update
    TableProxy( const std::string& s, bool ro=true ):
        tab( s, ((ro==true)?(Table::Old):(Table::Update)) ),
        ra( tab )
    {
    }

    // attempt to re-open the table for updating.
    // The .reopenRW() method is called and 
    // the doc. sais that an exception is thrown
    // if that fails. We catch the exception and
    // transform that into a false return value.
    bool update( void ) {
        bool   rv( true );
        try {
            tab.reopenRW();
        }
        catch( ... ) {
            rv = false;
        }
        return rv;
    }
    unsigned int ncol( void ) const
    {
        return tab.tableDesc().ncolumn();
    }

    unsigned int nrow( void ) const {
        return tab.nrow();
    }

    // all the columnnames as array of strings
    PyObject* colnames( void ) const {
        Vector<String>  cn( tab.tableDesc().columnNames() );
        return arraytransformer<warp::ToPyStr>()((const Array<String>&)cn);
    }
   
    // get all the values in row 'idx' in the table
    // supports python-style indexing where negative
    // indices count backwards from the end, ie, backward
    // from the last row of the table...
    PyObject* getRow( int r ) {
        static  boost::python::to_python_value<RecordInterface const&>  warper;
        if( r<0 ) {
            r += (int)tab.nrow();
        }
        if( r<0 || r>=(int)tab.nrow() ) {
            PyErr_SetString(PyExc_StopIteration, "");
            return NULL;
        }
        return warper(ra.get(r));
    }

    ColumnProxy* getColumn( const std::string& colnm )
    {
        return new ColumnProxy(tab, colnm);
    }

    const casa::String& tableName( void ) const {
        return tab.tableName();
    }

    PyObject* keywordSet( void ) 
    {
        return boost::python::to_python_value<RecordInterface const&>()( tab.keywordSet() );
    }

    TableProxy  query( const std::string& taql ) {
        return TableProxy( tableCommand(String("select from $1 where ")+String(taql), tab) );

    }

    // query with select clause
    TableProxy query( const std::string& cols,
                      const std::string& where )
    {
        String  qry("select ");

        if( cols.size() )
            qry += String(cols);
        qry += " from $1 ";
        if( where.size() )
            qry += String("where ")+String(where);
        return TableProxy( ::tableCommand(qry, tab) );
    }

    // taql on this table. your expression
    // is sent throught TaQL as-is
    TableProxy taql( const std::string& s ) {
        return TableProxy( ::tableCommand(String(s), tab) );
    }

    Table       tab;
    ROTableRow  ra;
};

// wrapper so as we do not have to bother about overloaded
// tableCommand() fn's. This is just the one taking a single
// string as arg, the full TaQL query
TableProxy  tc_wrapper(const std::string& s) {
    return TableProxy( ::tableCommand(String(s)) );
}

// Create function pointers to both versions of TableProxy::query
// in order to help Boost decide which to select...
TableProxy (TableProxy::*q1)(const std::string&) = &TableProxy::query;
TableProxy (TableProxy::*q2)(const std::string&, const std::string&) = &TableProxy::query;


PyObject* shapetst( PyObject* o ) {
    Py_INCREF(o);
    IPosition           shp( sequencesupport::Shaper::get_shape(o) );
    Py_DECREF(o);
    return arraytransformer<>()( shp.asVector() );
}


//
// The casa module
// 
BOOST_PYTHON_MODULE(pycasa)
{
    // define a table
    class_<TableProxy>("table")
        .def(init<>())
        .def(init<std::string>())
        .def(init<std::string,bool>())
        .def("tableName", &TableProxy::tableName,
                          return_value_policy<copy_const_reference>())
        .def("__len__", &TableProxy::nrow)
        .def("ncol", &TableProxy::ncol)
        .def("getcol", &TableProxy::getColumn,
                       return_internal_reference<>())
        .def("__getitem__", &TableProxy::getRow)
        .def("query", q1)
        .def("query", q2)
        .def("taql", &TableProxy::taql)
        .def("colnames", &TableProxy::colnames)
        .def("keywordSet", &TableProxy::keywordSet)
        .def("update", &TableProxy::update)
        ;
        
    class_<ColumnProxy>("column", no_init)
        .def("__str__", &stringifier<ColumnProxy>::asString )
        .def("__getitem__", &ColumnProxy::getCell )
        .def("__setitem__", &ColumnProxy::putCell )
        .def("get", &ColumnProxy::get)
        .def("__len__",&ColumnProxy::length)
        .def("keywordSet", &ColumnProxy::keywordSet)
        ;

    // define the global taql function
    def("taql", &tc_wrapper);

    // and test
//    def("shapetst", &shapetst);

    // Do register type conversions
    CasaStringFromPythonString();
}

//
// Transform a recordinterface based field into a PyDict
// Need this wrapper since it must be defined *after*
// the boost::python::to_python_value<RecordInterface> is defined
//
PyObject* make_pydict_fromsubrecord(casa::RecordInterface const& r, Int num) {
    // retrieve pointer to the subrecord
    const casa::RecordInterface& srec = *((RecordInterface*)(r.get_pointer(num, TpRecord)));

    // and warp it into a PyObject
    return boost::python::to_python_value<casa::RecordInterface const&>()(srec);
}

// Define this only after the class_<TableProxy> is defined 
// such that we can actually give that back
PyObject* make_pyobject_fromsubtable(casa::RecordInterface const& r, Int num) {
    // retrieve pointer to the subtable
    const casa::TableKeyword& trec = *((const TableKeyword*)(r.get_pointer(num, TpTable)));
    // Prepare return value, it's going to be one of our own defined
    // python classes (the tableproxy thingy) and we get a reference to the
    // held C++ object of type 'TableProxy'. Then we set the value to the
    // new table and return the python object
    object  rv( new TableProxy(trec.table()) );
    
    incref( rv.ptr() );
    return rv.ptr();
}
